/*
 *	Armadillo-2x0 GPIO Driver
 *	Copyright (C) 2005-2006 Atmark Techno, Inc.
 */

#include <linux/config.h>
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <asm/io.h>

#include <asm/arch/platform.h>
#include <asm/arch/armadillo2x0_gpio.h>

#define AUTHOR      "Atmark Techno, Inc."
#define DESCRIPTION "Armadillo-2x0 GPIO driver"
#define LICENSE     "GPL v2"

#define GPIO_MINOR   (185)
#define GPIO_DEVICE  "gpio"

#define VERSION     GPIO_DEVICE ": " DESCRIPTION ", (C) 2005-2006 " AUTHOR

#undef DEBUG
#if defined(DEBUG)
#define KDEBUG(args...) printk("gpio: " args)
#else
#define KDEBUG(args...)
#endif

struct wait_handler{
	wait_queue_head_t queue;
	unsigned long wait_list;
	unsigned long int_list;

	struct wait_handler *prev;
	struct wait_handler *next;
};
static struct wait_handler *wait_list_head = NULL;

struct gpio_port_reg {
	unsigned long ddr;
	unsigned long dr;
	unsigned long inten;
	unsigned long inttype1;
	unsigned long inttype2;
	unsigned long eoi;
	unsigned long db;
	unsigned long rawintsts;
	unsigned long intsts;
};

static struct gpio_port_reg porta = {
	.ddr       = GPIO_PADDR,
	.dr        = GPIO_PADR,
	.inten     = GPIO_AINTEN,
	.inttype1  = GPIO_AINTTYPE1,
	.inttype2  = GPIO_AINTTYPE2,
	.eoi       = GPIO_AEOI,
	.db        = GPIO_ADB,
	.rawintsts = GPIO_RAWINTSTATUSA,
	.intsts    = GPIO_INTSTATUSA,
};

static struct gpio_port_reg portb = {
	.ddr       = GPIO_PBDDR,
	.dr        = GPIO_PBDR,
	.inten     = GPIO_BINTEN,
	.inttype1  = GPIO_BINTTYPE1,
	.inttype2  = GPIO_BINTTYPE2,
	.eoi       = GPIO_BEOI,
	.db        = GPIO_BDB,
	.rawintsts = GPIO_RAWINTSTATUSB,
	.intsts    = GPIO_INTSTATUSB,
};

static struct gpio_port_reg portc = {
	.ddr       = GPIO_PCDDR,
	.dr        = GPIO_PCDR,
};

static struct gpio_port_reg portd = {
	.ddr       = GPIO_PDDDR,
	.dr        = GPIO_PDDR,
};

struct gpio_info {
	unsigned long no;
	struct gpio_port_reg *port;
	unsigned long bit;
};

static struct gpio_info gpio[] = {
	{GPIO0, &porta, BIT(4)},
	{GPIO1, &porta, BIT(5)},
	{GPIO2, &porta, BIT(6)},
	{GPIO3, &porta, BIT(7)},
	{GPIO4, &portb, BIT(0)},
	{GPIO5, &portb, BIT(1)},
	{GPIO6, &portb, BIT(2)},
	{GPIO7, &portb, BIT(3)},
	{GPIO8, &portd, BIT(0)},
	{GPIO9, &portd, BIT(1)},
	{GPIO10,&portd, BIT(2)},
	{GPIO11,&portd, BIT(3)},
	{GPIO12,&portc, BIT(0)},
	{GPIO13,&portc, BIT(1)},
	{GPIO14,&portc, BIT(2)},
	{GPIO15,&portc, BIT(3)},
};

static int gpio_num = 0;

#define change_reg_bit(REG,BIT,VALUE)       \
({                                          \
	unsigned tmp;                       \
	tmp = (inl(REG) & ~BIT);            \
	outl(tmp | (VALUE ? BIT : 0), REG); \
})

#define get_reg_bit(REG,BIT)                \
	inl(REG) & BIT

#if defined(DEBUG)
static void debug_wait_handle(struct wait_handler *head){
	int i;
	struct wait_handler *tmp;

	KDEBUG("debug_wait_handle");
	KDEBUG("display wait_handle_list\n");
	if(head == NULL){
		printk("list not found\n");
		return;
	}
	
	tmp = head;
	for(i=0;;i++){
		if(tmp ==NULL) break;
		printk("chain[%2d]: 0x%08lx\n",i,(unsigned long)tmp);
		printk("     prev: 0x%08lx\n",(unsigned long)tmp->prev);
		printk("     next: 0x%08lx\n",(unsigned long)tmp->next);
		printk("\n");
		tmp = tmp->next;
	}
}
#else
#define debug_wait_handle(head)
#endif

static struct wait_handler *create_wait_handle(unsigned long wait_list){
	struct wait_handler *tmp;
	struct wait_handler *handle;
	KDEBUG("create_wait_handle\n");
	handle = (struct wait_handler *)kmalloc(sizeof(struct wait_handler), 
						GFP_KERNEL);
	if(handle == NULL) return NULL;
	
	init_waitqueue_head(&handle->queue);
	handle->wait_list = wait_list;
	handle->int_list = 0;
	handle->next = NULL;
	
	if(wait_list_head){
		tmp = wait_list_head;    
		while(tmp->next){
			tmp = tmp->next;
		}
		tmp->next = handle;
		handle->prev= tmp;
	}else{
		wait_list_head = handle;
		handle->prev=NULL;
	}
	
	debug_wait_handle(wait_list_head);
	return handle;
}

static void destroy_wait_handle(struct wait_handler **handle){
	struct wait_handler *tmp;
	KDEBUG("destroy_wait_handle\n");
	tmp = wait_list_head;
	
	if(*handle == NULL) return;
	
	while(tmp != *handle){
		tmp = tmp->next;
		if(tmp ==NULL) return;
	}
	
	if(tmp->prev)
		tmp->prev->next = tmp->next;
	if(tmp->next)
		tmp->next->prev = tmp->prev;

	if(tmp->prev == NULL)
		wait_list_head = tmp->next;

	kfree(*handle);
	
	*handle = NULL;
	
	debug_wait_handle(wait_list_head);
	return;
}

static irqreturn_t gpio_irq_handler(int irq, void *dev_id,
				    struct pt_regs *regs)
{
	struct wait_handler *handle;
	unsigned long int_sts_a;
	unsigned long int_sts_b;
	int i;
	
	KDEBUG("gpio_irq_handler\n");
	
	int_sts_a = inl(GPIO_INTSTATUSA);
	int_sts_b = inl(GPIO_INTSTATUSB);
	
	if(wait_list_head != NULL){

		handle = wait_list_head;

		do{
			for(i=0;i<gpio_num;i++){
				if(handle->wait_list & gpio[i].no &&
				   gpio[i].port->inten){
					if(get_reg_bit(gpio[i].port->intsts,gpio[i].bit)){
						handle->int_list = gpio[i].no;
			
						wake_up_interruptible(&handle->queue);
					}
				}
			}
			handle = handle->next;
		}while(handle);
		
	}
	
	outl(int_sts_a, GPIO_AEOI);
	outl(int_sts_b, GPIO_BEOI);
	
	return IRQ_HANDLED;
}

static int gpio_open(struct inode *inode, struct file *file)
{
	KDEBUG("gpio_open\n");
	return 0;
}

static int gpio_release(struct inode *inode, struct file *file)
{
	KDEBUG("gpio_release\n");
	return 0;
}

static int set_param(struct gpio_param *param){
	int i;
	KDEBUG("set_param\n");
	
	switch(param->mode){
	case MODE_OUTPUT:
		KDEBUG("MODE_OUTPUT\n");
		for(i=0;i<gpio_num;i++){
			if(param->no == gpio[i].no){
				//DDR <= OUTPUT
				change_reg_bit(gpio[i].port->ddr, gpio[i].bit, 1);

				//DR <= value
				change_reg_bit(gpio[i].port->dr, gpio[i].bit, param->data.o.value);
				
				KDEBUG("ditect GPIO%d (%s)\n",i,(param->data.o.value ? "ON" : "OFF"));
				return 0;
			}
		}
		return -1;
		
	case MODE_INPUT:
		KDEBUG("MODE_INPUT\n");
		for(i=0;i<gpio_num;i++){
			if(param->no == gpio[i].no){
				//DDR <= INPUT
				change_reg_bit(gpio[i].port->ddr, gpio[i].bit, 0);

				if(param->data.i.int_enable &&
				   gpio[i].port->inten){
					//INTTYPE1,INTTYPE2 <= interrupt type
					switch(param->data.i.int_type & 0x03){
					case TYPE_LOW_LEVEL:
						change_reg_bit(gpio[i].port->inttype1, gpio[i].bit, 0);
						change_reg_bit(gpio[i].port->inttype2, gpio[i].bit, 0);
						break;
					case TYPE_HIGH_LEVEL:
						change_reg_bit(gpio[i].port->inttype1, gpio[i].bit, 0);
						change_reg_bit(gpio[i].port->inttype2, gpio[i].bit, 1);
						break;
					case TYPE_FALLING_EDGE:
						change_reg_bit(gpio[i].port->inttype1, gpio[i].bit, 1);
						change_reg_bit(gpio[i].port->inttype2, gpio[i].bit, 0);
						break;
					case TYPE_RISING_EDGE:
						change_reg_bit(gpio[i].port->inttype1, gpio[i].bit, 1);
						change_reg_bit(gpio[i].port->inttype2, gpio[i].bit, 1);
						break;
					}

					//DB <= debounce
					if(param->data.i.int_type & TYPE_DEBOUNCE){
						change_reg_bit(gpio[i].port->db, gpio[i].bit, 1);
					}

					//EOI <= clear
					change_reg_bit(gpio[i].port->eoi, gpio[i].bit, 1);
					
					//INTEN <= enable
					change_reg_bit(gpio[i].port->inten, gpio[i].bit, 1);
					
				}else{
					change_reg_bit(gpio[i].port->eoi, gpio[i].bit, 1);
					change_reg_bit(gpio[i].port->inten, gpio[i].bit, 0);
					change_reg_bit(gpio[i].port->inttype1, gpio[i].bit, 0);
					change_reg_bit(gpio[i].port->inttype2, gpio[i].bit, 0);
					change_reg_bit(gpio[i].port->db, gpio[i].bit, 0);
				}
				
				KDEBUG("ditect GPIO%d\n",i);
				return 0;
			}
		}
		return -1;
		break;
	default: return -1;
	}
	return 0;
}

static int get_param(struct gpio_param *param){
	int i;
	unsigned long io;
	
	KDEBUG("get_param\n");
	
	for(i=0;i<gpio_num;i++){
		if(param->no == gpio[i].no){ 
			io = get_reg_bit(gpio[i].port->ddr, gpio[i].bit);
	    
			if(io){
				//MODE_OUTPUT => mode
				param->mode = MODE_OUTPUT;

				//DR => value
				param->data.o.value = \
				  (get_reg_bit(gpio[i].port->dr, gpio[i].bit) ? 1 : 0); 
				
				KDEBUG("ditect GPIO%d (%s)\n",i,(param->data.o.value ? "ON" : "OFF"));
				return 0;
			}else{
				//MODE_INPUT => mode
				param->mode = MODE_INPUT;
			  
				//DR => value
				param->data.i.value = \
				  (get_reg_bit(gpio[i].port->dr, gpio[i].bit) ? 1 : 0);
				
				if(gpio[i].port->inten){
				  //INTEN => inten
				  param->data.i.int_enable = \
				    (get_reg_bit(gpio[i].port->inten, gpio[i].bit) ? 1 : 0);
				  
				  if(param->data.i.int_enable){
					//INTTYPE1,INTTYPE2 => interrupt type
					param->data.i.int_type = \
					  (get_reg_bit(gpio[i].port->inttype1, gpio[i].bit) ? BIT(1) : 0);
					param->data.i.int_type |= \
					  (get_reg_bit(gpio[i].port->inttype2, gpio[i].bit) ? BIT(0) : 0);
					
					//DB => debounce
					param->data.i.int_type |= \
					  (get_reg_bit(gpio[i].port->db, gpio[i].bit) ? BIT(2) : 0);
				  }
				  else {
				    param->data.i.int_enable = 0;
				    param->data.i.int_type = 0;
				  }
				}
				
				KDEBUG("ditect GPIO%d\n",i);
				return 0;
			}
		}
	}
	return -EINVAL;
}

static int gpio_ioctl(struct inode *inode, struct file *file,
		      unsigned int cmd,unsigned long arg)
{
	struct wait_handler *wait_handle;
	struct wait_param *wait_param;
	struct gpio_param *param_list;
	struct gpio_param *param;
	int ret;
	int i;
	
	KDEBUG("gpio_ioctl\n");
	
	switch(cmd){
	case PARAM_SET:
		KDEBUG("PARAM_SET\n");
		param_list = (struct gpio_param *)arg;
		param=param_list;
		
		for(i=0;i<gpio_num;i++){
		  
			if(param==NULL) break;

			ret = set_param(param);
			if(ret) return -EINVAL;
			
			param = param->next;
		}
		break;
		
	case PARAM_GET:
		KDEBUG("PARAM_GET\n");
		param_list = (struct gpio_param *)arg;
		param=param_list;
		
		for(i=0;i<gpio_num;i++){
		  
			if(param==NULL) break;
			
			ret = get_param(param);
			if(ret) return -EINVAL;
			
			param = param->next;
		}
		break;
		
	case INTERRUPT_WAIT:
		wait_param = (struct wait_param *)arg;
		KDEBUG("INTERRUPT_WAIT\n");
		
		wait_handle = create_wait_handle(wait_param->list);
		if(wait_handle == NULL) return -EBUSY;
		
		if(wait_param->timeout){
			interruptible_sleep_on_timeout
			  (&wait_handle->queue,
			   //msec->jiffies
			   (wait_param->timeout * HZ + 999) / 1000);
		}else{
			interruptible_sleep_on(&wait_handle->queue);
		}
		
		wait_param->list = wait_handle->int_list;
		
		destroy_wait_handle(&wait_handle);
		
		break;
	default:
		return -EINVAL;
	}
	
	return 0;
}

static struct file_operations gpio_fops = {
	.owner   = THIS_MODULE,
	.open    = gpio_open,
	.release = gpio_release,
	.ioctl   = gpio_ioctl,
};

static struct miscdevice gpio_dev = {
	.minor = GPIO_MINOR,
	.name  = GPIO_DEVICE,
	.fops  = &gpio_fops,
};

static int __init gpio_init(void)
{
	int ret;
	KDEBUG("gpio_init\n");

	switch(get_board_id()){
	case BOARD_ID_A210:
	case BOARD_ID_A210C:
		gpio_num = GPIO_NUM_A210;
		break;
	case BOARD_ID_A220:
	case BOARD_ID_A230:
	case BOARD_ID_A240:
		gpio_num = GPIO_NUM;
		break;
	}

	if(misc_register(&gpio_dev)){
		return -ENODEV;
	}

	printk(KERN_INFO VERSION "\n");
	
	ret = request_irq(IRQ_GPIO,gpio_irq_handler,SA_INTERRUPT,"gpio",NULL);
	if(ret) return ret;
	
	return 0;
}

static void __exit gpio_exit(void)
{
	KDEBUG("gpio_exit\n");
	free_irq(IRQ_GPIO,NULL);
	misc_deregister(&gpio_dev);
}

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE(LICENSE);

module_init(gpio_init);
module_exit(gpio_exit);
